/*
* Dieses kleine Projekt dient dem Zweck, den Umgang mit einem
* STM32 Cortex M4 Mikrocontroller zu illustrieren. Es wird hierbei
* bewusst auf die Verwendung der blichen Libraries verzichtet, um
* mglichst wenig von der eigentlichen Vorgehensweise zu "verdecken".
* Es geht hierbei nicht darum, mit mglichst wenig Aufwand "das Ding
* ans Laufen zu kriegen", sondern ein grundlegendes Verstndnis ber
* die Funktionsweise des M4 zu erlangen.
*
* Bentigte Dokumente:
* [1] RM0090 Reference Manual (DM00031020.pdf)
* [2] STM32F405xx/STM32F407xx datasheet (DM00037051.pdf)
*
* Falls das STM32F4 discovery board genutzt wird:
* [3] UM1472 User Manual (DM00039084.pdf)
*
* Autor: J. Kerdels
* Lizenz: CC BY 3.0
*/

// Reset and Clock Control
#include "rcc.h"

// Methoden fr das STM32F4 discovery board
#include "discovery.h"

// Beispiele fr das STM32F4 discovery board
#include "discovery_ex.h"
#include "libfoo/stm32f4xx.h"
// In der Datei discovery_ex.c befinden sich #defines, die - wenn
// einkommentiert - das entsprechende Beispiel auswhlen

//uint8_t jbuff[320*240] __attribute__ ((seciton (".lram128k")));

/*
 * fixed point Julia set in 3.29 format
 * Zoom faktor in powers of 2 :)
 *
 */
void GenerateJulia_fpu(uint16_t size_x, uint16_t size_y, uint16_t offset_x, uint16_t offset_y, uint16_t zoom, uint8_t * buffer)
{
uint32_t tmp1, tmp2;
uint32_t num_real, num_img;
uint32_t radius, fofs_y, fofs_x, fzoom;
const uint32_t rlimit = 0x80000000; // 4.0

uint8_t i;
uint16_t x,y;
#define ITERATION 128
#define IMG_CONSTANT  0x00083125 // 000.0_0000_0000_1000_0011_0001_0010_0110 ~0.001f
#define REAL_CONSTANT 0x091ed841 // 000.0_1001_0001_1110_1011_1000_0100_0001 0.284999998286 ~.285f

	fofs_y = (float)offset_y;
	fofs_x = (float)offset_x;
	fzoom = (float) zoom;
	for (y=0; y<size_y; y++)
	{
		for (x=0; x<size_x; x++)
		{
			num_real = (float)y - fofs_y;
			num_real = num_real / zoom;
			num_img = (float)x - fofs_x;
			num_img = num_img / zoom;
			i=0;
			radius = 0.0f;
			while ((i<ITERATION-1) && (radius < rlimit))
			{
				tmp1 = num_real * num_real;
				tmp2 = num_img * num_img;
				num_img = 2*num_real*num_img + IMG_CONSTANT;
				num_real = tmp1 - tmp2 + REAL_CONSTANT;
				radius = tmp1 + tmp2;
				i++;
			}
			/* Store the value in the buffer */
			buffer[x+y*size_x] = i;
		}
	}
}

void GenerateJulia_fpu(uint16_t size_x, uint16_t size_y, uint16_t offset_x, uint16_t offset_y, uint16_t zoom, uint8_t * buffer)
{
float tmp1, tmp2;
float num_real, num_img;
float radius, fofs_y, fofs_x, fzoom;
const float rlimit = 4.0f;
uint8_t i;
uint16_t x,y;
#define ITERATION 128
#define IMG_CONSTANT 0.001f
#define REAL_CONSTANT 0.285f

	fofs_y = (float)offset_y;
	fofs_x = (float)offset_x;
	fzoom = (float) zoom;
	for (y=0; y<size_y; y++)
	{
		for (x=0; x<size_x; x++)
		{
			num_real = (float)y - fofs_y;
			num_real = num_real / zoom;
			num_img = (float)x - fofs_x;
			num_img = num_img / zoom;
			i=0;
			radius = 0.0f;
			while ((i<ITERATION-1) && (radius < rlimit))
			{
				tmp1 = num_real * num_real;
				tmp2 = num_img * num_img;
				num_img = 2*num_real*num_img + IMG_CONSTANT;
				num_real = tmp1 - tmp2 + REAL_CONSTANT;
				radius = tmp1 + tmp2;
				i++;
			}
			/* Store the value in the buffer */
			buffer[x+y*size_x] = i;
		}
	}
}


int main(void)
{
    // Zu Beginn der Verarbeitung muss die gewnschte Taktung der
    // einzelnen Komponenten des Systems eingestellt werden. Dies
    // wird hier von der Methode rcc_init() in der Datei rcc.c
    // erledigt.
    rcc_init();
    
    // Das STM32F4 discovery board bringt einige externe Komponenten
    // mit sich. Die Methode discovery_basic_init() richtet zunchst
    // die Schnittstellen der einfacheren Komponenten (LEDs und Taster)
    // des STM32F4 ein.
    discovery_basic_init();

    for (;;)
	{
		GPIOD->ODR ^= 0x0000F000;
		GenerateJulia_fpu(320, 240, 160, 120, 120, (uint8_t *) 0x20000000);
		GPIOD->ODR ^= 0x0000F000;
		GenerateJulia_fpu(320, 240, 160, 120, 120, (uint8_t *) 0x20000000);
	}
	
	//----------------------------------------------------------------------
    
    // In diesem einfachen Beispiel werden die 4 LEDs des discovery boards
    // eingeschaltet, wenn der User-Button des boards gedrckt wird.
    led_and_button_example();

    //----------------------------------------------------------------------
    
    // Dieses Beispiel zeigt, wie man die Grundkonfiguration eines Timers
    // durchfhrt und wie man den Timer auf sehr einfach Weise nutzen kann.
    led_and_timer_example();
    
    //----------------------------------------------------------------------
    
    // In diesem Beispiel soll die Verwendung eines Interrupts erlutert
    // werden.
    timer_irq_example();

    //----------------------------------------------------------------------

    // In diesem Beispiel soll die Verwendung eines Timers zur PWM-Generierung
    // erlutert werden.
    pwm_led_example();

    //----------------------------------------------------------------------

    // Das folgende Beispiel demonstriert, wie man mithilfe des DMA controllers
    // gnzlich ohne Unterbrechung der Hauptprogrammschleife die LEDs weich
    // ein- und ausblenden lassen kann.
    dma_pwm_led_example();


    return 0;
}